# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

"""
Allows you to load YAML files with filename and lineno information in
sets, lists and dicts.
"""

from __future__ import absolute_import
import collections
import io
import types
import yaml

__all__ = ['load']


def _yaml_type_factory(name, base):
  """
  Generates a subclass for the specified *base* class where its constructor
  accepts a filename and line number as the first two arguments.
  """

  class _YamlType(base):

    def __init__(self, filename, lineno, *args, **kwargs):
      super(_YamlType, self).__init__(*args, **kwargs)
      self.filename = filename
      self.lineno = lineno

    def __repr__(self):
      return '{}({})'.format(type(self).__name__, super(_YamlType, self).__repr__())

  _YamlType.__name__ = name
  if hasattr(_YamlType, '__qualname__'):
    _YamlType.__qualname__ = name

  return _YamlType


YamlSet = _yaml_type_factory('YamlSet', set)
YamlList = _yaml_type_factory('YamlList', list)
YamlDict = _yaml_type_factory('YamlDict', dict)


class YamlLineNumberLoaderMixin(object):

  def __init__(self, filename):
    self.__filename = filename

    # Make sure the yaml_constructors are hooked up.
    self.yaml_constructors = self.yaml_constructors.copy()
    for key, value in self.yaml_constructors.items():
      if hasattr(type(self), value.__name__):
        self.yaml_constructors[key] = getattr(type(self), value.__name__)

  def compose_node(self, parent, index):
    # Propagate current line numbers to the Yaml AST nodes.
    line = self.line
    node = super(YamlLineNumberLoaderMixin, self).compose_node(parent, index)
    node.__line__ = line + 1
    return node

  def construct_mapping(self, node, deep=False):
    value = super(YamlLineNumberLoaderMixin, self).construct_mapping(node, deep=deep)
    return YamlDict(self.__filename, node.__line__, value)

  def construct_sequence(self, node, deep=False):
    value = super(YamlLineNumberLoaderMixin, self).construct_sequence(node, deep=deep)
    return YamlList(self.__filename, node.__line__, value)

  def construct_pairs(self, node, deep=False):
    value = super(YamlLineNumberLoaderMixin, self).construct_pairs(node, deep=deep)
    return YamlList(self.__filename, node.__line__, value)

  def construct_yaml_set(self, node):
    data = YamlSet(self.__filename, node.__line__)
    yield data
    value = self.construct_mapping(node)
    data.update(value)

  def construct_yaml_seq(self, node):
    data = YamlList(self.__filename, node.__line__)
    yield data
    data.extend(self.construct_sequence(node))

  def construct_yaml_map(self, node):
    data = YamlDict(self.__filename, node.__line__)
    yield data
    value = self.construct_mapping(node)
    data.update(value)


def load(_data=None, *args, **kwargs):
  """
  Loads YAML data from a file, adding line numbers to every type of collection
  returned from the loader (sets, lists and dicts). The specified YAML loader
  will be mixed with the #YamlLineNumberLoaderMixin class. The default loader
  is #yaml.SafeLoader.
  """

  loader_cls = kwargs.pop('Loader', yaml.SafeLoader)
  filename = kwargs.pop('filename', None)

  if _data is None and filename is not None:
    with io.open(filename, encoding=kwargs.pop('encoding', None)) as fp:
      _data = fp.read()

  class _YamlLoader(YamlLineNumberLoaderMixin, loader_cls):
    def __init__(self, *args, **kwargs):
      loader_cls.__init__(self, *args, **kwargs)
      YamlLineNumberLoaderMixin.__init__(self, filename)

  loader = _YamlLoader(_data, *args, **kwargs)
  return loader.get_single_data()



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import logging
import os
import threading
import sys

try:
  import watchdog
except ImportError:
  watchdog = None
  MaybeFileSystemEventHandler = object
else:
  from watchdog.observers import Observer
  from watchdog.events import FileSystemEventHandler as MaybeFileSystemEventHandler


class ConfigReloaderTask(object):
  """
  A helper class to help reloading config files in the background.
  """

  def __init__(self, filename, load_config_function, logger=None):
    self.filename = filename
    self.load_config_function = load_config_function
    self._num_reloads = 0
    self._observer = None
    self._callbacks = []
    self._logger = logger or logging.getLogger('ConfigReloaderTask'.format(filename))
    self._value = None
    self._reload_lock = threading.Lock()
    self._data_lock = threading.Condition()

  def get(self):  # type: () -> Any
    """
    Returns the currently loaded configuration. If the configuration was never
    loaded, it will be loaded for the first time. Note that this can return None
    if the config could not be loaded.
    """

    with self._data_lock:
      if self._num_reloads > 0:
        return self._value
    return self.reload()

  def reload(self):
    """
    Reloads the configuration. Two concurrent calls of this method will result
    in one thread waiting for the other to finish and only one actual reload
    to occur. Sequential calls will result in multiple actual reloads.
    """

    # If another call to reload() is currently in progress, we wait until that
    # is finished and we can retrieve the new configuration value.
    if self._reload_lock.locked():
      self._logger.debug('filename=%r, reloading already in progress...', self.filename)
      with self._data_lock:
        self._data_lock.wait()
        return self._value

    # Indicate that a reload is in progress.
    with self._reload_lock:
      try:
        config = self.load_config_function(self.filename)
      except:
        self._logger.exception('filename=%r, error during reload.', self.filename)
        with self._data_lock:
          self._data_lock.notify_all()
      else:
        with self._data_lock:
          self._num_reloads += 1
          self._value = config
          # Notify all other calls stuck in reload() waiting for this one to finish
          # that the new value can be read.
          self._data_lock.notify_all()

        # Invoke callbacks for the changed config.
        for callback in self._callbacks:
          try:
            callback(config)
          except:
            self._logger.exception('filename=%r, error during callback %r.', self.filename, callback)

        return config

  def reload_callback(self, callback):
    """
    Registers a callback that is invoked when the config was reloaded. The new
    configuration will be passed as sole argument to the callback.
    """

    self._callbacks.append(callback)

  def start(self):
    if self._observer:  # TODO: Check if the observer is running.
      raise RuntimeError('ConfigReloaderTask is already running.')
    self._logger.debug('filename=%r, starting reloader thread.', self.filename)
    self._observer = WatchdogFileObserver(self.filename, self._file_modified)
    self._observer.start()

  def stop(self):
    if self._observer:
      self._observer.stop()
      self._observer.join()
      self._observer = None

  def _file_modified(self, event):
    self._logger.debug('filename=%r, received file modified event.', self.filename)
    self.reload()


class WatchdogFileObserver(MaybeFileSystemEventHandler):

  def __init__(self, filename, callback):
    if watchdog is None:
      raise RuntimeError('watchdog module is not available')
    self.filename = os.path.normpath(os.path.abspath(filename))
    self.callback = callback
    self.observer = Observer()
    self.observer.schedule(self, path=os.path.dirname(self.filename), recursive=False)

  def start(self):
    self.observer.start()

  def stop(self):
    self.observer.stop()

  def join(self):
    self.observer.join()

  def on_modified(self, event):
    if event.src_path == self.filename:
      self.callback(event)



# -*- coding: utf8 -*-
# Copyright (c) 2020 Niklas Rosenstein
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from nr.collections import abc
from nr.pylang.utils.classdef import make_singleton
from typing import List, Union
import copy
import os
import re
import six

#: This is a special singleton return by #parse_accessor_string() to represent
#: a wildcard index.
WILDCARD = make_singleton('WILDCARD')


def parse_accessor_string(s):  # type: (str) -> List[Union[str, int, WILDCARD]]
  """
  Parses an a string consisting of a sequence of access instructions and returns it
  as a list of keys and indices.

  >>> from nr.config.plugins import parse_accessor_string
  >>> parse_accessor_string('a.b[0].c')
  ['a', 'b', 0, 'c']
  >>> parse_accessor_string('a[*].b')
  ['a', WILDCARD, 'b']
  """

  result = []
  for part in s.split('.'):
    match = re.match(r'(.*)\[(\d+|\*)\]', part)
    if match:
      result.append(match.group(1))
      result.append(WILDCARD if match.group(2) == '*' else int(match.group(2)))
    else:
      result.append(part)

  return result


def resolve_accessor_list(l, data):  # type: (List[Union[str, int, WILDCARD]], Any) -> Any
  """
  Resolves a list of access instructions on *data* and returns the value.
  """

  it = iter(l)
  for item in it:
    if item == WILDCARD:
      # TODO: Verify that *item* is a sequence?
      subl = list(it)
      data = [resolve_accessor_list(subl, v) for v in data]
      break
    else:
      data = data[item]

  return data


class Vars(object):
  """
  A preprocessor plugin that substitutes strings of the form `{{variableName}}`
  with the actual value that they reference. References that encompass the full
  string can be expanded into values of other types. References as part of a
  longer strings will be rendered in that part of the string.
  """

  _DEFAULT_REGEX = r'\{\{([^}]+)\}\}'

  def __init__(self, iterable=None, resolve=None, regex=None, safe=False):
    """
    :param iterable: Used to create a new #dict() if specified. Cannot be used at
      the same time as *resolve*.
    :param resolve: Resolve the text inside `{{...}}` to return a replacement
      value. If not specified, the default implementation will be used (using
      #parse_accessor_string() and #resolve_accessor_list().
    :param regex: The regex to match variables. The regex must have exactly one
      group defined (the name of the variable).
    :param safe: If True, KeyError or IndexError exceptions raised by *resolver*
      will be caught and the substitution will not occur.
    """

    if iterable is not None:
      if resolve is not None:
        raise TypeError('iterable and resolve cannot be specified together')
      data = dict(iterable)
      resolver = lambda key: resolve_accessor_list(parse_accessor_string(key), self._data)
    else:
      data = None

    self._data = data
    self._resolver = resolver
    self._regex = re.compile(regex or self._DEFAULT_REGEX)
    self._safe = safe

  def __sub(self, match):
    " Private. Passed to #re.sub() for replacements. "

    key = match.group(1)
    try:
      return self._resolver(key)
    except (KeyError, IndexError):
      if self._safe:
        return match.group(0)
      raise

  def __call__(self, data):
    if isinstance(data, six.string_types):
      match = self._regex.match(data)
      if match and match.group(0) == data:
        return self.__sub(match)
      else:
        return self._regex.sub(lambda m: str(self.__sub(m)), data)
    return data


class Include(object):
  """
  Replaces strings of the form `{{!include <FILENAME>}}` with the contents of
  the actual file, optionally loaded with the specified *load_func*.
  """

  def __init__(self, base_dir, load_func=None):
    self.base_dir = base_dir
    self.load_func = load_func

  def __call__(self, data):
    if not isinstance(data, six.string_types):
      return data
    match = re.match(r'{{\s*!include\s+(.+?)\s+}}$', data)
    if not match:
      return data
    filename = os.path.join(self.base_dir, match.group(1))
    with open(filename) as fp:
      if self.load_func:
        return self.load_func(fp)
      return fp.read()


class Envvars(object):
  """
  Replaces references of `{{ env VARIABLE_NAME }}` with the actual environment
  variable value.
  """

  def __init__(self, resolve=None):
    if resolve is None:
      resolve = lambda k: os.environ[k]
    self._resolve = resolve

  def __call__(self, data):
    if not isinstance(data, six.string_types):
      return data
    def sub(match):
      envname = match.group(1)
      return self._resolve(match.group(1))
    return re.sub(r'{{\s*env\s+([\w_\d]+)\s*}}', sub, data)


class Processor(object):
  """
  Encapsulates plugins and some options and then allows processing nested
  data recursively.
  """

  def __init__(self, plugins=(), mutate=False, keep_type=True):
    self._plugins = list(plugins)
    self._mutate = mutate
    self._keep_type = keep_type

  def add_plugin(self, plugin):
    self._plugins.append(plugin)

  def __call__(self, data):
    """
    Process the specified *data* with the list of *plugins* and return the
    result. Handles strings, mappings and sequences. If the *mutate* is True,
    mappings and sequences will be assumed mutable. If *keep_type* is True, the
    function will attempt to keep the same type of the mapping or sequence,
    otherwise dicts or lists are returned.
    """

    for plugin in self._plugins:
      data = plugin(data)

    if isinstance(data, abc.Mapping):
      if self._mutate and isinstance(data, abc.MutableMapping):
        for key in data:
          data[key] = self(data[key])
      else:
        cls = type(data) if self._keep_type else dict
        data = cls((k, self(data[k])) for k in data)
    elif isinstance(data, abc.Sequence) and not isinstance(data, six.string_types):
      if self._mutate and isinstance(data, abc.MutableSequence):
        for index, value in enumerate(data):
          data[index] = self(value)
      else:
        cls = type(data) if self._keep_type else list
        data = cls(self(v) for v in data)

    return data


def process_config(data, plugins):
  return Processor(plugins)(data)


def merge_config(a, b, path='$'):
  if isinstance(a, abc.Mapping):
    if not isinstance(b, abc.Mapping):
      raise ValueError('Unable to merge incompatible types "{}" and "{}" at {}'
                       .format(type(a).__name__, type(b).__name__, path))
    a = copy.copy(a)
    for key in b:
      if key in a:
        a[key] = merge_config(a[key], b[key], path + '.' + key)
      else:
        a[key] = b[key]
    return a
  return b
